import time

from arcor2.data.common import Joint, Pose, StrEnum
from arcor2_dobot.dobot import Dobot, DobotException
from arcor2_dobot.dobot_api import DobotApiException


class Joints(StrEnum):

    J1: str = "dobot_m1_axis_2_joint"
    J2: str = "dobot_m1_axis_3_joint"
    J3: str = "dobot_m1_z_axis_joint"
    J4: str = "dobot_m1_axis_4_joint"


class DobotM1(Dobot):
    def __init__(self, pose: Pose, port: str = "/dev/dobot", simulator: bool = False) -> None:
        super(DobotM1, self).__init__(pose, port, simulator)

        if self.simulator:
            self._joint_values = [Joint(Joints.J1, 0), Joint(Joints.J2, 0), Joint(Joints.J3, 0), Joint(Joints.J4, 0)]

    def _handle_pose_in(self, pose: Pose) -> None:

        pose.position.x -= 0.11
        pose.position.y -= 0.0
        pose.position.z += 0.01

    def _handle_pose_out(self, pose: Pose) -> None:

        pose.position.x += 0.11
        pose.position.y += 0.0
        pose.position.z -= 0.01

    def robot_joints(self, include_gripper: bool = False) -> list[Joint]:

        if self.simulator:
            return self._joint_values

        joints = self._dobot.get_pose().joints.in_radians()
        j3 = self._dobot.get_pose().joints.j3 / 1000
        return [
            Joint(Joints.J1, joints.j1),
            Joint(Joints.J2, joints.j2),
            Joint(Joints.J3, j3),
            Joint(Joints.J4, joints.j4),
        ]

    def suck(self) -> None:

        if self.simulator:
            return

        try:
            self._dobot.set_io(17, False)  # suck
            self._dobot.set_io(18, False)  # on
        except DobotApiException as e:
            raise DobotException("Suck failed.") from e

    def release(self, blow_out_sec: float = 0.1) -> None:

        if self.simulator:
            time.sleep(blow_out_sec)
            return

        try:
            if blow_out_sec > 0:
                self._dobot.set_io(17, True)  # blow
                self._dobot.set_io(18, False)  # on
                time.sleep(blow_out_sec)

            self._dobot.set_io(18, True)  # off

        except DobotApiException as e:
            raise DobotException("Release failed.") from e



from arcor2.flask import FlaskException, General, WebApiErrorFactory
from arcor2_dobot import __name__ as package_name


class DobotException(FlaskException):
    service = package_name


class DobotGeneral(FlaskException):
    description = General.description


class NotFound(DobotException):
    description = "Occurs when something is not found"


class StartError(DobotException):
    description = "Occurs when start condition is not met"


WebApiError = WebApiErrorFactory.get_class(DobotGeneral, NotFound, StartError)



import math
import time
from abc import ABCMeta, abstractmethod

import quaternion

import arcor2.transformations as tr
from arcor2.data.common import Joint, Orientation, Pose, StrEnum
from arcor2.exceptions import Arcor2NotImplemented
from arcor2.helpers import NonBlockingLock
from arcor2.object_types.abstract import RobotException
from arcor2_dobot.dobot_api import MODE_PTP, DobotApi, DobotApiException

# TODO jogging


class DobotException(RobotException):
    pass


class MoveType(StrEnum):

    JUMP: str = "JUMP"
    JOINTS: str = "JOINTS"
    LINEAR: str = "LINEAR"


MOVE_TYPE_MAPPING = {
    MoveType.JUMP: MODE_PTP.JUMP_XYZ,
    MoveType.JOINTS: MODE_PTP.MOVJ_XYZ,
    MoveType.LINEAR: MODE_PTP.MOVL_XYZ,
}


class Dobot(metaclass=ABCMeta):

    ROTATE_EEF = Orientation.from_rotation_vector(y=math.pi)
    UNROTATE_EEF = ROTATE_EEF.inversed()

    def __init__(self, pose: Pose, port: str = "/dev/dobot", simulator: bool = False) -> None:

        self.pose = pose
        self.simulator = simulator
        self._move_lock = NonBlockingLock()

        if not self.simulator:

            try:
                self._dobot = DobotApi(port)
            except DobotApiException as e:
                raise DobotApiException("Could not connect to the robot.") from e

        else:
            self._joint_values: list[Joint] = []  # has to be set to some initial value in derived classes
            self._hand_teaching = False

    def alarms_to_exception(self) -> None:

        try:
            alarms = self._dobot.get_alarms()
        except DobotApiException as e:
            raise DobotApiException("Failed to get alarms.") from e

        if alarms:
            raise DobotApiException(f"Alarm(s): {','.join([alarm.name for alarm in alarms])}.")

    def cleanup(self) -> None:

        self.conveyor_speed(0)
        self.release()

        if not self.simulator:
            self._dobot.close()

    def conveyor_speed(self, speed: float, direction: int = 1) -> None:

        if not self.simulator:
            self._dobot.conveyor_belt(speed, direction)

    def conveyor_distance(self, speed: float, distance: float, direction: int = 1) -> None:
        self._dobot.conveyor_belt_distance(speed, distance, direction)

    @property
    def hand_teaching_mode(self) -> bool:

        if self.simulator:
            return self._hand_teaching

        return self._dobot.get_hht_trig_output()

    @hand_teaching_mode.setter
    def hand_teaching_mode(self, value: bool) -> None:

        if self.simulator:
            self._hand_teaching = value
            return

        self._dobot.set_hht_trig_output(value)

    def _inverse_kinematics(self, pose: Pose) -> list[Joint]:
        raise Arcor2NotImplemented()

    def inverse_kinematics(self, pose: Pose) -> list[Joint]:
        raise Arcor2NotImplemented()

    def forward_kinematics(self, joints: list[Joint]) -> Pose:
        raise Arcor2NotImplemented()

    def _handle_pose_out(self, pose: Pose) -> None:
        """This is called (only for a real robot) from `get_end_effector_pose`
        so derived classes can do custom changes to the pose.

        :param pose:
        :return:
        """

        pass

    def _handle_pose_in(self, pose: Pose) -> None:
        """This is called (only for a real robot) from `move` so derived
        classes can do custom changes to the pose.

        :param pose:
        :return:
        """

        pass

    def _check_orientation(self, pose: Pose) -> None:

        unrotated = self.UNROTATE_EEF * pose.orientation

        eps = 1e-6

        if abs(unrotated.x) > eps or abs(unrotated.y) > eps:
            raise DobotApiException("Impossible orientation.")

    def get_end_effector_pose(self) -> Pose:

        if self.simulator:
            return self.forward_kinematics(self.robot_joints())

        try:
            pos = self._dobot.get_pose()  # in mm
        except DobotApiException as e:
            raise DobotException("Failed to get pose.") from e

        p = Pose()
        p.position.x = pos.position.x / 1000.0
        p.position.y = pos.position.y / 1000.0
        p.position.z = pos.position.z / 1000.0
        p.orientation = self.ROTATE_EEF * Orientation.from_rotation_vector(z=math.radians(pos.position.r))

        self._handle_pose_out(p)
        return tr.make_pose_abs(self.pose, p)

    def home(self) -> None:
        """Run the homing procedure."""

        with self._move_lock:

            if self.simulator:
                time.sleep(2.0)
                return

            try:
                self._dobot.clear_alarms()
                self._dobot.wait_for_cmd(self._dobot.home())
            except DobotApiException as e:
                raise DobotException("Homing failed.") from e

        self.alarms_to_exception()

    def move(self, pose: Pose, move_type: MoveType, velocity: float = 50.0, acceleration: float = 50.0) -> None:
        """Moves the robot's end-effector to a specific pose.

        :param pose: Target pose.
        :move_type: Move type.
        :param velocity: Speed of move (percent).
        :param acceleration: Acceleration of move (percent).
        :return:
        """

        if not (0.0 <= velocity <= 100.0):
            raise DobotException("Invalid velocity.")

        if not (0.0 <= acceleration <= 100.0):
            raise DobotException("Invalid acceleration.")

        with self._move_lock:

            rp = tr.make_pose_rel(self.pose, pose)

            # prevent Dobot from moving when given an unreachable goal
            try:
                jv = self._inverse_kinematics(rp)
            except Arcor2NotImplemented:  # TODO remove this once M1 has IK
                pass

            if self.simulator:
                self._joint_values = jv
                time.sleep((100.0 - velocity) * 0.05)
                return

            self._handle_pose_in(rp)

            try:
                self._dobot.clear_alarms()
                unrotated = self.UNROTATE_EEF * rp.orientation
                rotation = math.degrees(quaternion.as_rotation_vector(unrotated.as_quaternion())[2])
                self._dobot.speed(velocity, acceleration)

                self._dobot.wait_for_cmd(
                    self._dobot.move_to(
                        rp.position.x * 1000.0,
                        rp.position.y * 1000.0,
                        rp.position.z * 1000.0,
                        rotation,
                        MOVE_TYPE_MAPPING[move_type],
                    )
                )
            except DobotApiException as e:
                raise DobotException("Move failed.") from e

        self.alarms_to_exception()

    @abstractmethod
    def suck(self) -> None:
        pass

    @abstractmethod
    def release(self) -> None:
        pass

    @abstractmethod
    def robot_joints(self) -> list[Joint]:
        pass



import math

import quaternion

import arcor2.transformations as tr
from arcor2.data.common import Joint, Orientation, Pose, Position, StrEnum
from arcor2_dobot.dobot import Dobot, DobotException
from arcor2_dobot.dobot_api import DobotApiException


class Joints(StrEnum):

    J1: str = "magician_joint_1"
    J2: str = "magician_joint_2"
    J3: str = "magician_joint_3"
    J4: str = "magician_joint_4"
    J5: str = "magician_joint_5"


class DobotMagician(Dobot):

    # Dimensions in meters (according to URDF)
    link_2_length = 0.135
    link_3_length = 0.147
    link_4_length = 0.06
    end_effector_length = 0.06

    def __init__(self, pose: Pose, port: str = "/dev/dobot", simulator: bool = False) -> None:
        super(DobotMagician, self).__init__(pose, port, simulator)

        if self.simulator:
            self._joint_values = [
                Joint(Joints.J1, -0.038),
                Joint(Joints.J2, 0.341),
                Joint(Joints.J3, 0.3632),
                Joint(Joints.J4, -0.704),
                Joint(Joints.J5, -0.568),
            ]

    def _handle_pose_in(self, pose: Pose) -> None:

        base_angle = math.atan2(pose.position.y, pose.position.x)
        pose.position.x -= self.link_4_length * math.cos(base_angle)
        pose.position.y -= self.link_4_length * math.sin(base_angle)
        pose.position.z += self.end_effector_length

    def _handle_pose_out(self, pose: Pose) -> None:

        base_angle = math.atan2(pose.position.y, pose.position.x)
        pose.position.x += self.link_4_length * math.cos(base_angle)
        pose.position.y += self.link_4_length * math.sin(base_angle)
        pose.position.z -= self.end_effector_length

    # TODO joint4/5
    valid_ranges: dict[Joints, tuple[float, float]] = {
        Joints.J1: (-2, 2),
        Joints.J2: (-0.1, 1.46),
        Joints.J3: (-0.95, 1.15),
    }

    def validate_joints(self, joints: list[Joint]) -> None:

        for joint in joints:
            if joint.name in self.valid_ranges:
                vrange = self.valid_ranges[Joints(joint.name)]
                if vrange[0] > joint.value < vrange[1]:
                    raise DobotException(
                        "Value {:.3f} for joint {:s} is out of the range of {:s}.".format(
                            joint.value, joint.name, str(vrange)
                        )
                    )

    def _inverse_kinematics(self, pose: Pose) -> list[Joint]:
        """Computes inverse kinematics.

        Works with robot-relative pose.

        Inspired by DobotKinematics.py from open-dobot project and DobotInverseKinematics.py from BenchBot.

        :param pose: IK target pose (relative to robot)
        :return: Inverse kinematics
        """

        self._check_orientation(pose)

        # TODO this is probably not working properly (use similar solution as in _check_orientation?)
        _, _, yaw = quaternion.as_euler_angles(pose.orientation.as_quaternion())

        x = pose.position.x
        y = pose.position.y
        z = pose.position.z + self.end_effector_length

        # pre-compute distances
        # radial position of end effector in the x-y plane
        r = math.sqrt(math.pow(x, 2) + math.pow(y, 2))
        rho_sq = pow(r - self.link_4_length, 2) + pow(z, 2)
        rho = math.sqrt(rho_sq)  # distance b/w the ends of the links joined at the elbow

        l2_sq = self.link_2_length**2
        l3_sq = self.link_3_length**2

        # law of cosines
        try:
            alpha = math.acos((l2_sq + rho_sq - l3_sq) / (2.0 * self.link_2_length * rho))
            gamma = math.acos((l2_sq + l3_sq - rho_sq) / (2.0 * self.link_2_length * self.link_3_length))
        except ValueError:
            raise DobotException("Failed to compute IK.")

        beta = math.atan2(z, r - self.link_4_length)

        # joint angles
        baseAngle = math.atan2(y, x)
        rearAngle = math.pi / 2 - beta - alpha
        frontAngle = math.pi / 2 - gamma

        joints = [
            Joint(Joints.J1, baseAngle),
            Joint(Joints.J2, rearAngle),
            Joint(Joints.J3, frontAngle),
            Joint(Joints.J4, -rearAngle - frontAngle),
            Joint(Joints.J5, yaw - baseAngle),
        ]
        self.validate_joints(joints)
        return joints

    def inverse_kinematics(self, pose: Pose) -> list[Joint]:
        """Computes inverse kinematics.

        Works with absolute pose.

        :param end_effector_id: IK target pose end-effector
        :param pose: IK target pose
        :param start_joints: IK start joints (not supported)
        :param avoid_collisions: Return non-collision IK result if true (not supported)
        :return: Inverse kinematics
        """

        return self._inverse_kinematics(tr.make_pose_rel(self.pose, pose))

    def forward_kinematics(self, joints: list[Joint]) -> Pose:
        """Computes forward kinematics.

        Outputs absolute pose.

        Inspired by DobotKinematics.py from open-dobot project.

        :param end_effector_id: Target end effector name
        :param joints: Input joint values
        :return: Pose of the given end effector
        """

        self.validate_joints(joints)

        j1 = joints[0].value
        j2 = joints[1].value
        j3 = joints[2].value
        sj = j2 + j3

        radius = (
            self.link_2_length * math.cos(j2 - math.pi / 2) + self.link_3_length * math.cos(sj) + self.link_4_length
        )

        x = radius * math.cos(j1)
        y = radius * math.sin(j1)

        z = (
            self.link_2_length * math.cos(j2)
            + self.link_3_length * math.cos(sj + math.pi / 2)
            - self.end_effector_length
        )

        pose = Pose(
            Position(x, y, z),
            Orientation.from_quaternion(quaternion.from_euler_angles(0, math.pi, joints[-1].value + j1)),
        )

        if __debug__:
            self._check_orientation(pose)

        return tr.make_pose_abs(self.pose, pose)

    def robot_joints(self, include_gripper: bool = False) -> list[Joint]:

        if self.simulator:
            return self._joint_values

        joints = self._dobot.get_pose().joints.in_radians()
        return [
            Joint(Joints.J1, joints.j1),
            Joint(Joints.J2, joints.j2),
            Joint(Joints.J3, joints.j3 - joints.j2),
            Joint(Joints.J4, -joints.j3),
            Joint(Joints.J5, joints.j4),
        ]

    def suck(self) -> None:

        if self.simulator:
            return

        try:
            self._dobot.wait_for_cmd(self._dobot.suck(True))
        except DobotApiException as e:
            raise DobotException("Suck failed.") from e

    def release(self) -> None:

        if self.simulator:
            return

        try:
            self._dobot.wait_for_cmd(self._dobot.suck(False))
        except DobotApiException as e:
            raise DobotException("Release failed.") from e




# DO NOT EDIT THIS FILE -- AUTOGENERATED BY PANTS
# Target: src/python/arcor2_dobot:arcor2_dobot_dist

from setuptools import setup

setup(**{
    'author': 'Robo@FIT',
    'author_email': 'imaterna@fit.vut.cz',
    'classifiers': [
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
        'Programming Language :: Python :: 3.9',
        'Natural Language :: English',
        'Topic :: Scientific/Engineering',
    ],
    'description': 'ARCOR2 Dobot service',
    'entry_points': {
        'console_scripts': [
            'arcor2_dobot = arcor2_dobot.scripts.dobot:main',
        ],
    },
    'install_requires': (
        'Flask~=2.2.2',
        'arcor2~=0.24.2',
        'numpy-quaternion[numba,scipy]~=2022.4.2',
        'numpy~=1.23.2',
        'pyserial~=3.5',
    ),
    'license': 'LGPL',
    'long_description': """# arcor2_dobot

## Environment variables

- `ARCOR2_DOBOT_URL=http://0.0.0.0:5018` - by default, the service listens on port 5018.
- `ARCOR2_DOBOT_MOCK=1` - the service will start in a mock (simulator) mode.
 `ARCOR2_DOBOT_DEBUG=1` - turns on debug logging.
# Changelog

The format is based on [Keep a Changelog](https://keepachangelog.com/en/1.0.0/),

## [0.5.0] - 2022-09-02

### Changed
- **BREAKING**: Implement new error handling flow. Error codes of **every** endpoint were replaced with error 
  types as described in swagger documentation.
- Updated API description.
- Model and port are now set using environment variables.

### Added
- New exception classes

## [0.4.1] - 2022-02-17

### Added

- Support for safe movements.
  - Connection to Scene Service Mock needed.
  - New parameter `safe` for `PUT /eef/pose`.
  - For JOINT and JUMP movements: add intermediate points with higher z, if there might be a collision.
  - For LINEAR: raise an exception if there might be collision.
- Debug logging can be turned on by adding `-d` or setting `ARCOR2_DOBOT_DEBUG`.

### Fixed

- In mock-mode, `cleanup` was throwing `AttributeError`.

## [0.4.0] - 2022-01-25

### Added

- Support for conveyor belt.

### Changed

- Suction is switched off during `stop`.

## [0.3.0] - 2021-10-25

### Changed
 - New environment variable `ARCOR2_DOBOT_MOCK` to start the service in a mock mode.
 - Classes updated to match `Robot` (new argument `include_gripper`). 
 - Handle `DobotApi` exceptions to produce nice error messages.
 - Use HTTP code 204 instead of 200 where it makes sense.

## [0.2.3] - 2021-05-21

### Fixed
 - EEF orientation.
 - State controller paths.

## [0.2.2] - 2021-03-17

### Fixed
 - Suction offset for M1 adjusted

## [0.2.1] - 2021-03-15

### Fixed
 - Dobot axis Z is in mm
 - Added fixed offset for suction

## [0.2.0] - 2021-03-03

### Changed
- Integration of `pydobot` library (own modified fork).
- API for hand teaching mode.

## [0.1.0] - 2021-02-08

### Changed
- Initial release of the Dobot service.""",
    'long_description_content_type': 'text/markdown',
    'name': 'arcor2_dobot',
    'namespace_packages': (
    ),
    'package_data': {
        'arcor2_dobot': (
            'VERSION',
            'py.typed',
        ),
    },
    'packages': (
        'arcor2_dobot',
        'arcor2_dobot.scripts',
    ),
    'python_requires': '==3.9.*',
    'version': '0.5.0',
})



#!/usr/bin/env python3

import argparse
import copy
import json
import logging
import os
from functools import wraps
from typing import Optional

from flask import Response, jsonify, request

from arcor2 import env
from arcor2.clients import scene_service
from arcor2.data.common import Joint, Pose, StrEnum
from arcor2.data.scene import LineCheck
from arcor2.flask import RespT, create_app, run_app
from arcor2.helpers import port_from_url
from arcor2.logging import get_logger
from arcor2_dobot import version
from arcor2_dobot.dobot import Dobot, DobotApiException, MoveType
from arcor2_dobot.exceptions import DobotGeneral, NotFound, StartError, WebApiError
from arcor2_dobot.m1 import DobotM1
from arcor2_dobot.magician import DobotMagician

logger = get_logger(__name__)


class DobotModels(StrEnum):
    MAGICIAN: str = "magician"
    M1: str = "m1"


URL = os.getenv("ARCOR2_DOBOT_URL", "http://localhost:5018")
DOBOT_PORT = os.getenv("ARCOR2_DOBOT_PORT", "/dev/dobot")
DOBOT_MODEL = DobotModels(os.getenv("ARCOR2_DOBOT_MODEL", DobotModels.MAGICIAN))

SERVICE_NAME = f"Dobot Web API ({DOBOT_MODEL})"

dobot_model_mapping: dict[DobotModels, type[Dobot]] = {DobotModels.MAGICIAN: DobotMagician, DobotModels.M1: DobotM1}

assert set(dobot_model_mapping.keys()) == DobotModels.set()


app = create_app(__name__)

_dobot: Optional[Dobot] = None
_mock = False


def started() -> bool:

    return _dobot is not None


def requires_started(f):
    @wraps(f)
    def wrapped(*args, **kwargs):
        if not started():
            raise StartError("Not started")
        return f(*args, **kwargs)

    return wrapped


@app.route("/state/start", methods=["PUT"])
def put_start() -> RespT:
    """Start the robot.
    ---
    put:
        description: Start the robot.
        tags:
           - State
        requestBody:
              content:
                application/json:
                  schema:
                    $ref: Pose
        responses:
            204:
              description: Ok
            500:
              description: "Error types: **General**, **DobotGeneral**, **StartError**."
              content:
                application/json:
                  schema:
                    $ref: WebApiError
    """

    if started():
        raise StartError("Already started.")

    if not isinstance(request.json, dict):
        raise DobotGeneral("Body should be a JSON dict containing Pose.")

    pose = Pose.from_dict(request.json)

    global _dobot

    _dobot = dobot_model_mapping[DOBOT_MODEL](pose, DOBOT_PORT, _mock)

    return Response(status=204)


@app.route("/state/stop", methods=["PUT"])
@requires_started
def put_stop() -> RespT:
    """Stop the robot.
    ---
    put:
        description: Stop the robot.
        tags:
           - State
        responses:
            204:
              description: Ok
            500:
              description: "Error types: **General**, **StartError**."
              content:
                application/json:
                  schema:
                    $ref: WebApiError
    """

    global _dobot
    assert _dobot is not None
    _dobot.cleanup()
    _dobot = None
    return Response(status=204)


@app.route("/state/started", methods=["GET"])
def get_started() -> RespT:
    """Get the current state.
    ---
    get:
        description: Get the current state.
        tags:
           - State
        responses:
            200:
              description: Ok
              content:
                application/json:
                    schema:
                        type: boolean
            500:
              description: "Error types: **General**."
              content:
                application/json:
                  schema:
                    $ref: WebApiError
    """

    return jsonify(started())


@app.route("/conveyor/speed", methods=["PUT"])
@requires_started
def put_conveyor_speed() -> RespT:
    """Set the conveyor belt speed.
    ---
    put:
        description: Set the conveyor belt speed.
        tags:
           - Conveyor Belt
        parameters:
            - name: velocity
              in: query
              schema:
                type: number
                default: 50.0
                format: float
                minimum: 0
                maximum: 100
            - in: query
              name: direction
              schema:
                type: string
                default: forward
                enum:
                    - forward
                    - backwards
              description: Direction
        responses:
            204:
              description: Ok
            500:
              description: "Error types: **General**, **StartError**."
              content:
                application/json:
                  schema:
                    $ref: WebApiError
    """

    speed = float(request.args.get("velocity", default=50.0))
    direction = request.args.get("direction", default="forward")

    assert _dobot is not None
    _dobot.conveyor_speed(speed, 1 if direction == "forward" else -1)
    return Response(status=204)


@app.route("/conveyor/distance", methods=["PUT"])
@requires_started
def put_conveyor_distance() -> RespT:
    """Set the conveyor belt distance.
    ---
    put:
        description: Set the conveyor belt distance.
        tags:
           - Conveyor Belt
        parameters:
            - name: velocity
              in: query
              schema:
                type: number
                default: 50.0
                format: float
                minimum: 0
                maximum: 100
            - name: distance
              in: query
              schema:
                type: number
                default: 0.1
                format: float
                minimum: 0
            - in: query
              name: direction
              schema:
                type: string
                default: forward
                enum:
                    - forward
                    - backwards
              description: Direction
        responses:
            204:
              description: Ok
            500:
              description: "Error types: **General**, **StartError**."
              content:
                application/json:
                  schema:
                    $ref: WebApiError
    """

    speed = float(request.args.get("velocity", default=50.0))
    direction = request.args.get("direction", default="forward")
    distance = float(request.args.get("distance", default=0.1))

    assert _dobot is not None
    _dobot.conveyor_distance(speed, distance * 1000, 1 if direction == "forward" else -1)
    return Response(status=204)


@app.route("/eef/pose", methods=["GET"])
@requires_started
def get_eef_pose() -> RespT:
    """Get the EEF pose.
    ---
    get:
        description: Get the EEF pose.
        tags:
           - Robot
        responses:
            200:
              description: Ok
              content:
                application/json:
                    schema:
                        $ref: Pose
            500:
              description: "Error types: **General**, **StartError**."
              content:
                application/json:
                  schema:
                    $ref: WebApiError
    """

    assert _dobot is not None
    return jsonify(_dobot.get_end_effector_pose()), 200


@app.route("/eef/pose", methods=["PUT"])
@requires_started
def put_eef_pose() -> RespT:
    """Set the EEF pose.
    ---
    put:
        description: Set the EEF pose.
        tags:
           - Robot
        parameters:
            - in: query
              name: moveType
              schema:
                type: string
                enum:
                    - JUMP
                    - LINEAR
                    - JOINTS
              required: true
              description: Move type
            - name: velocity
              in: query
              schema:
                type: number
                format: float
                minimum: 0
                maximum: 100
            - name: acceleration
              in: query
              schema:
                type: number
                format: float
                minimum: 0
                maximum: 100
            - in: query
              name: safe
              schema:
                type: boolean
                default: false
        requestBody:
              content:
                application/json:
                  schema:
                    $ref: Pose
        responses:
            200:
              description: Ok
            500:
              description: "Error types: **General**, **DobotGeneral**, **StartError**, **NotFound**."
              content:
                application/json:
                  schema:
                    $ref: WebApiError
    """

    assert _dobot is not None

    if not isinstance(request.json, dict):
        raise DobotGeneral("Body should be a JSON dict containing Pose.")

    pose = Pose.from_dict(request.json)
    move_type = MoveType(request.args.get("moveType", "jump"))
    velocity = float(request.args.get("velocity", default=50.0))
    acceleration = float(request.args.get("acceleration", default=50.0))
    safe = request.args.get("safe") == "true"

    if safe:
        cp = _dobot.get_end_effector_pose()

        ip1 = copy.deepcopy(cp)
        ip2 = copy.deepcopy(pose)

        for _attempt in range(20):
            res = scene_service.line_check(LineCheck(ip1.position, ip2.position))

            if res.safe:
                break

            if move_type == MoveType.LINEAR:
                raise DobotGeneral("There might be a collision.")

            ip1.position.z += 0.01
            ip2.position.z += 0.01

        else:
            raise NotFound("Can't find safe path.")

        logger.debug(f"Collision avoidance attempts: {_attempt}")

        if _attempt > 0:
            _dobot.move(ip1, move_type, velocity, acceleration)
            _dobot.move(ip2, move_type, velocity, acceleration)

    _dobot.move(pose, move_type, velocity, acceleration)
    return Response(status=204)


@app.route("/home", methods=["PUT"])
@requires_started
def put_home() -> RespT:
    """Get the current state.
    ---
    put:
        description: Get the current state.
        tags:
           - Robot
        responses:
            204:
              description: Ok
            500:
              description: "Error types: **General**, **StartError**."
              content:
                application/json:
                  schema:
                    $ref: WebApiError
    """

    assert _dobot is not None
    _dobot.home()
    return Response(status=204)


@app.route("/hand_teaching", methods=["GET"])
@requires_started
def get_hand_teaching() -> RespT:
    """Get hand teaching status.
    ---
    get:
        description: Get hand teaching status.
        tags:
           - Robot
        responses:
            200:
              description: Ok
              content:
                application/json:
                    schema:
                        type: boolean
            500:
              description: "Error types: **General**, **StartError**."
              content:
                application/json:
                  schema:
                    $ref: WebApiError
    """

    assert _dobot
    return jsonify(_dobot.hand_teaching_mode)


@app.route("/hand_teaching", methods=["PUT"])
@requires_started
def put_hand_teaching() -> RespT:
    """Set hand teaching status.
    ---
    put:
        description: Set hand teaching status.
        tags:
           - Robot
        parameters:
            - in: query
              name: enabled
              schema:
                type: boolean
        responses:
            204:
              description: Ok
            500:
              description: "Error types: **General**, **StartError**."
              content:
                application/json:
                  schema:
                    $ref: WebApiError
    """

    assert _dobot
    _dobot.hand_teaching_mode = request.args.get("enabled") == "true"
    return Response(status=204)


@app.route("/suck", methods=["PUT"])
@requires_started
def put_suck() -> RespT:
    """Get the current state.
    ---
    put:
        description: Get the current state.
        tags:
           - Robot
        responses:
            204:
              description: Ok
            500:
              description: "Error types: **General**, **StartError**."
              content:
                application/json:
                  schema:
                    $ref: WebApiError
    """

    assert _dobot is not None
    _dobot.suck()
    return Response(status=204)


@app.route("/release", methods=["PUT"])
@requires_started
def put_release() -> RespT:
    """Get the current state.
    ---
    put:
        description: Get the current state.
        tags:
           - Robot
        responses:
            204:
              description: Ok
            500:
              description: "Error types: **General**, **StartError**."
              content:
                application/json:
                  schema:
                    $ref: WebApiError
    """

    assert _dobot is not None
    _dobot.release()
    return Response(status=204)


@app.route("/joints", methods=["GET"])
@requires_started
def get_joints() -> RespT:
    """Get the current state.
    ---
    get:
        description: Get the current state.
        tags:
           - Robot
        responses:
            200:
              description: Ok
              content:
                application/json:
                    schema:
                        type: array
                        items:
                            $ref: Joint
            500:
              description: "Error types: **General**, **StartError**."
              content:
                application/json:
                  schema:
                    $ref: WebApiError
    """

    assert _dobot is not None
    return jsonify(_dobot.robot_joints())


@app.route("/ik", methods=["PUT"])
@requires_started
def put_ik() -> RespT:
    """Get the current state.
    ---
    put:
        description: Get the current state.
        tags:
           - Robot
        requestBody:
              content:
                application/json:
                  schema:
                    $ref: Pose
        responses:
            200:
              description: Ok
              content:
                application/json:
                    schema:
                        type: array
                        items:
                            $ref: Joint
            500:
              description: "Error types: **General**, **DobotGeneral**, **StartError**."
              content:
                application/json:
                  schema:
                    $ref: WebApiError
    """

    assert _dobot is not None

    if not isinstance(request.json, dict):
        raise DobotGeneral("Body should be a JSON dict containing Pose.")

    pose = Pose.from_dict(request.json)
    return jsonify(_dobot.inverse_kinematics(pose))


@app.route("/fk", methods=["PUT"])
@requires_started
def put_fk() -> RespT:
    """Get the current state.
    ---
    put:
        description: Get the current state.
        tags:
           - Robot
        requestBody:
              content:
                application/json:
                  schema:
                    type: array
                    items:
                        $ref: Joint
        responses:
            200:
              description: Ok
              content:
                application/json:
                    schema:
                        $ref: Pose
            500:
              description: "Error types: **General**, **DobotGeneral**, **StartError**."
              content:
                application/json:
                  schema:
                    $ref: WebApiError
    """

    assert _dobot is not None

    if not isinstance(request.json, list):
        raise DobotGeneral("Body should be a JSON array containing joints.")

    joints = [Joint.from_dict(j) for j in request.json]
    return jsonify(_dobot.forward_kinematics(joints))


@app.errorhandler(DobotApiException)
def handle_dobot_exception(e: DobotApiException) -> tuple[str, int]:
    return json.dumps(DobotGeneral(str(e)).to_dict()), 500


@app.route("/health")
def health():
    """Healthcheck.
    ---
    get:
        tags:
            - Service
        description: Gets service health status.
        responses:
            204:
              description: The service is up and running.
    """
    return Response(status=204)


def main() -> None:

    parser = argparse.ArgumentParser(description=SERVICE_NAME)
    parser.add_argument("-s", "--swagger", action="store_true", default=False)
    parser.add_argument("-m", "--mock", action="store_true", default=env.get_bool("ARCOR2_DOBOT_MOCK"))

    parser.add_argument(
        "-d",
        "--debug",
        help="Set logging level to debug.",
        action="store_const",
        const=logging.DEBUG,
        default=logging.DEBUG if env.get_bool("ARCOR2_DOBOT_DEBUG") else logging.INFO,
    )

    args = parser.parse_args()
    logger.setLevel(args.debug)

    global _mock
    _mock = args.mock
    if _mock:
        logger.info("Starting as a mock!")

    if not args.swagger:
        scene_service.wait_for()

    run_app(
        app,
        SERVICE_NAME,
        version(),
        port_from_url(URL),
        [Pose, Joint, WebApiError],
        args.swagger,
        dependencies={"ARCOR2 Scene": "0.1.0"},
    )

    if _dobot:
        _dobot.cleanup()


if __name__ == "__main__":
    main()



import logging
import math
import struct
from collections import deque
from enum import IntEnum
from threading import RLock
from typing import NamedTuple, Optional

import numpy as np
import serial
from serial.tools import list_ports

MAX_QUEUE_LEN = 32

"""
This was originally https://github.com/luismesas/pydobot
"""


class Message:
    def __init__(self, b: Optional[bytes] = None) -> None:

        self.header: bytes = bytes([0xAA, 0xAA])
        self.len: int = 0x00
        self.ctrl: int = 0x00
        self.params: bytearray = bytearray([])
        self.checksum: Optional[int] = None

        if b:
            self.header = b[0:2]
            self.len = b[2]
            self.id = b[3]
            self.ctrl = b[4]
            self.params = bytearray(b[5:-1])
            self.checksum = b[-1:][0]

    def __repr__(self) -> str:
        return "Message()"

    def __str__(self) -> str:
        self.refresh()
        hex_header = " ".join("%02x" % b for b in self.header)
        hex_params = " ".join("%02x" % b for b in self.params)
        ret = "%s:%d:%d:%d:%s:%s" % (hex_header, self.len, self.id, self.ctrl, hex_params, self.checksum)
        return ret.upper()

    def refresh(self) -> None:
        if self.checksum is None:
            self.checksum = self.id + self.ctrl
            for param in self.params:
                self.checksum += param
            self.checksum = self.checksum % 256
            self.checksum = 2**8 - self.checksum
            self.checksum = self.checksum % 256
            self.len = 0x02 + len(self.params)

    def bytes(self) -> bytes:
        self.refresh()
        assert self.checksum is not None
        if len(self.params) > 0:
            command = bytearray([0xAA, 0xAA, self.len, self.id, self.ctrl])
            command.extend(self.params)
            command.append(self.checksum)
        else:
            command = bytearray([0xAA, 0xAA, self.len, self.id, self.ctrl, self.checksum])
        return command


class MODE_PTP(IntEnum):

    JUMP_XYZ = 0x00
    MOVJ_XYZ = 0x01
    MOVL_XYZ = 0x02
    JUMP_ANGLE = 0x03
    MOVJ_ANGLE = 0x04
    MOVL_ANGLE = 0x05
    MOVJ_INC = 0x06
    MOVL_INC = 0x07
    MOVJ_XYZ_INC = 0x08
    JUMP_MOVL_XYZ = 0x09


STEP_PER_CIRCLE = 360.0 / 1.8 * 10.0 * 16.0  # TODO how was this calculated?
MM_PER_CIRCLE = 3.1415926535898 * 36.0  # pi * diameter of c.b.'s shaft
HUNGARIAN_CONSTANT = 1.893  # to correct c.b. distance (obtained by measurement)


class DobotApiException(Exception):
    pass


class Position(NamedTuple):

    x: float
    y: float
    z: float
    r: float


class Joints(NamedTuple):

    j1: float
    j2: float
    j3: float
    j4: float

    def in_radians(self) -> "Joints":
        return Joints(math.radians(self.j1), math.radians(self.j2), math.radians(self.j3), math.radians(self.j4))


class Pose(NamedTuple):

    position: Position
    joints: Joints


class Alarm(IntEnum):

    COMMON_RESETTING = (0x00,)
    COMMON_UNDEFINED_INSTRUCTION = (0x01,)
    COMMON_FILE_SYSTEM = (0x02,)
    COMMON_MCU_FPGA_COMM = (0x03,)
    COMMON_ANGLE_SENSOR = 0x04

    PLAN_INV_SINGULARITY = (0x10,)
    PLAN_INV_CALC = (0x11,)
    PLAN_INV_LIMIT = (0x12,)  # !!!
    PLAN_PUSH_DATA_REPEAT = (0x13,)
    PLAN_ARC_INPUT_PARAM = (0x14,)
    PLAN_JUMP_PARAM = (0x15,)
    PLAN_LINE_HAND = (0x16,)
    PLAN_LINE_OUT_SPACE = (0x17,)
    PLAN_ARC_OUT_SPACE = (0x18,)
    PLAN_MOTIONTYPE = (0x19,)
    PLAN_SPEED_INPUT_PARAM = (0x1A,)
    PLAN_CP_CALC = (0x1B,)

    MOVE_INV_SINGULARITY = (0x20,)
    MOVE_INV_CALC = (0x21,)
    MOVE_INV_LIMIT = (0x22,)

    OVERSPEED_AXIS1 = (0x30,)
    OVERSPEED_AXIS2 = (0x31,)
    OVERSPEED_AXIS3 = (0x32,)
    OVERSPEED_AXIS4 = (0x33,)

    LIMIT_AXIS1_POS = (0x40,)
    LIMIT_AXIS1_NEG = (0x41,)
    LIMIT_AXIS2_POS = (0x42,)
    LIMIT_AXIS2_NEG = (0x43,)
    LIMIT_AXIS3_POS = (0x44,)
    LIMIT_AXIS3_NEG = (0x45,)
    LIMIT_AXIS4_POS = (0x46,)
    LIMIT_AXIS4_NEG = (0x47,)
    LIMIT_AXIS23_POS = (0x48,)
    LIMIT_AXIS23_NEG = 0x49

    LOSE_STEP_AXIS1 = (0x50,)
    LOSE_STEP_AXIS2 = 0x51
    LOSE_STEP_AXIS3 = 0x52
    LOSE_STEP_AXIS4 = 0x53

    OTHER_AXIS1_DRV_ALARM = (0x60,)
    OTHER_AXIS1_OVERFLOW = (0x61,)
    OTHER_AXIS1_FOLLOW = (0x62,)
    OTHER_AXIS2_DRV_ALARM = (0x63,)
    OTHER_AXIS2_OVERFLOW = (0x64,)
    OTHER_AXIS2_FOLLOW = (0x65,)
    OTHER_AXIS3_DRV_ALARM = (0x66,)
    OTHER_AXIS3_OVERFLOW = (0x67,)
    OTHER_AXIS3_FOLLOW = (0x68,)
    OTHER_AXIS4_DRV_ALARM = (0x69,)
    OTHER_AXIS4_OVERFLOW = (0x6A,)
    OTHER_AXIS4_FOLLOW = (0x6B,)

    MOTOR_REAR_ENCODER = (0x70,)
    MOTOR_REAR_TEMPERATURE_HIGH = 0x71
    MOTOR_REAR_TEMPERATURE_LOW = (0x72,)
    MOTOR_REAR_LOCK_CURRENT = (0x73,)
    MOTOR_REAR_BUSV_HIGH = (0x74,)
    MOTOR_REAR_BUSV_LOW = (0x75,)
    MOTOR_REAR_OVERHEAT = (0x76,)
    MOTOR_REAR_RUNAWAY = (0x77,)
    MOTOR_REAR_BATTERY_LOW = (0x78,)
    MOTOR_REAR_PHASE_SHORT = (0x79,)
    MOTOR_REAR_PHASE_WRONG = (0x7A,)
    MOTOR_REAR_LOST_SPEED = (0x7B,)
    MOTOR_REAR_NOT_STANDARDIZE = (0x7C,)
    ENCODER_REAR_NOT_STANDARDIZE = (0x7D,)
    MOTOR_REAR_CAN_BROKE = (0x7E,)

    MOTOR_FRONT_ENCODER = (0x80,)
    MOTOR_FRONT_TEMPERATURE_HIGH = (0x81,)
    MOTOR_FRONT_TEMPERATURE_LOW = (0x82,)
    MOTOR_FRONT_LOCK_CURRENT = (0x83,)
    MOTOR_FRONT_BUSV_HIGH = (0x84,)
    MOTOR_FRONT_BUSV_LOW = (0x85,)
    MOTOR_FRONT_OVERHEAT = (0x86,)
    MOTOR_FRONT_RUNAWAY = (0x87,)
    MOTOR_FRONT_BATTERY_LOW = (0x88,)
    MOTOR_FRONT_PHASE_SHORT = (0x89,)
    MOTOR_FRONT_PHASE_WRONG = (0x8A,)
    MOTOR_FRONT_LOST_SPEED = (0x8B,)
    MOTOR_FRONT_NOT_STANDARDIZE = (0x8C,)
    ENCODER_FRONT_NOT_STANDARDIZE = (0x8D,)
    MOTOR_FRONT_CAN_BROKE = (0x8E,)

    MOTOR_Z_ENCODER = (0x90,)
    MOTOR_Z_TEMPERATURE_HIGH = (0x91,)
    MOTOR_Z_TEMPERATURE_LOW = (0x92,)
    MOTOR_Z_LOCK_CURRENT = (0x93,)
    MOTOR_Z_BUSV_HIGH = (0x94,)
    MOTOR_Z_BUSV_LOW = (0x95,)
    MOTOR_Z_OVERHEAT = (0x96,)
    MOTOR_Z_RUNAWAY = (0x97,)
    MOTOR_Z_BATTERY_LOW = (0x98,)
    MOTOR_Z_PHASE_SHORT = (0x99,)
    MOTOR_Z_PHASE_WRONG = (0x9A,)
    MOTOR_Z_LOST_SPEED = (0x9B,)
    MOTOR_Z_NOT_STANDARDIZE = (0x9C,)
    ENCODER_Z_NOT_STANDARDIZE = (0x9D,)
    MOTOR_Z_CAN_BROKE = (0x9E,)

    MOTOR_R_ENCODER = (0xA0,)
    MOTOR_R_TEMPERATURE_HIGH = (0xA1,)
    MOTOR_R_TEMPERATURE_LOW = (0xA2,)
    MOTOR_R_LOCK_CURRENT = (0xA3,)
    MOTOR_R_BUSV_HIGH = (0xA4,)
    MOTOR_R_BUSV_LOW = (0xA5,)
    MOTOR_R_OVERHEAT = (0xA6,)
    MOTOR_R_RUNAWAY = (0xA7,)
    MOTOR_R_BATTERY_LOW = (0xA8,)
    MOTOR_R_PHASE_SHORT = 0xA9
    MOTOR_R_PHASE_WRONG = (0xAA,)
    MOTOR_R_LOST_SPEED = (0xAB,)
    MOTOR_R_NOT_STANDARDIZE = (0xAC,)
    ENCODER_R_NOT_STANDARDIZE = (0xAD,)
    MOTOR_R_CAN_BROKE = (0xAE,)

    MOTOR_ENDIO_IO = (0xB0,)
    MOTOR_ENDIO_RS485_WRONG = (0xB1,)
    MOTOR_ENDIO_CAN_BROKE = 0xB2


class DobotApi:
    def __init__(self, port: Optional[str] = None) -> None:

        self.logger = logging.Logger(__name__)
        self._lock = RLock()

        if port is None:
            # Find the serial port
            ports = list_ports.comports()
            for thing in ports:
                if thing.vid in (4292, 6790):
                    self.logger.debug(f"Found a com port to talk to DOBOT ({thing}).")
                    port = thing.device
                    break
            else:
                raise DobotApiException("Device not found!")

        try:
            self._ser = serial.Serial(
                port,
                baudrate=115200,
                parity=serial.PARITY_NONE,
                stopbits=serial.STOPBITS_ONE,
                bytesize=serial.EIGHTBITS,
            )
        except serial.serialutil.SerialException as e:
            raise DobotApiException from e

        self.logger.debug("%s open" % self._ser.name if self._ser.isOpen() else "failed to open serial port")

        self._set_queued_cmd_start_exec()
        self._set_queued_cmd_clear()
        self._set_ptp_joint_params(200, 200, 200, 200, 200, 200, 200, 200)
        self._set_ptp_coordinate_params(velocity=200, acceleration=200)
        self._set_ptp_jump_params(10, 200)
        self._set_ptp_common_params(velocity=100, acceleration=100)

        alarms = self.get_alarms()

        if alarms:
            self.logger.warning(f"Clearing alarms: {', '.join(map(str, alarms))}.")
            self.clear_alarms()

    def close(self) -> None:
        with self._lock:
            self._ser.close()
        self.logger.debug("%s closed" % self._ser.name)

    def _send_command(self, msg) -> Message:
        with self._lock:
            self._ser.reset_input_buffer()
            self._send_message(msg)
            msg = self._read_message()
        if msg is None:
            raise DobotApiException("No response!")
        return msg

    def _send_message(self, msg) -> None:

        self.logger.debug(msg)
        with self._lock:
            self._ser.write(msg.bytes())

    def _read_message(self) -> Optional[Message]:

        # Search for begin
        begin_found = False
        last_byte = None
        tries = 5
        while not begin_found and tries > 0:
            current_byte = ord(self._ser.read(1))
            if current_byte == 170:
                if last_byte == 170:
                    begin_found = True
            last_byte = current_byte
            tries = tries - 1
        if begin_found:
            payload_length = ord(self._ser.read(1))
            payload_checksum = self._ser.read(payload_length + 1)
            if len(payload_checksum) == payload_length + 1:
                b = bytearray([0xAA, 0xAA])
                b.extend(bytearray([payload_length]))
                b.extend(payload_checksum)
                msg = Message(b)
                self.logger.debug("Length", payload_length)
                self.logger.debug(payload_checksum)
                self.logger.debug("MessageID:", payload_checksum[0])
                self.logger.debug("<<", ":".join("{:02x}".format(x) for x in b))
                return msg
        return None

    def get_pose(self) -> Pose:
        msg = Message()
        msg.id = 10
        response = self._send_command(msg)

        return Pose(
            Position(
                struct.unpack_from("f", response.params, 0)[0],
                struct.unpack_from("f", response.params, 4)[0],
                struct.unpack_from("f", response.params, 8)[0],
                struct.unpack_from("f", response.params, 12)[0],
            ),
            Joints(
                struct.unpack_from("f", response.params, 16)[0],
                struct.unpack_from("f", response.params, 20)[0],
                struct.unpack_from("f", response.params, 24)[0],
                struct.unpack_from("f", response.params, 28)[0],
            ),
        )

    def get_alarms(self) -> set[Alarm]:

        msg = Message()
        msg.id = 20
        response = self._send_command(msg)  # 32 bytes

        ret: set[Alarm] = set()

        for idx in range(16):
            alarm_byte = struct.unpack_from("B", response.params, idx)[0]
            for alarm_index in [i for i in range(alarm_byte.bit_length()) if alarm_byte & (1 << i)]:
                ret.add(Alarm(idx * 8 + alarm_index))
        return ret

    def clear_alarms(self) -> None:

        msg = Message()
        msg.id = 20
        msg.ctrl = 0x01
        self._send_command(msg)  # empty response

    def _set_cp_cmd(self, x: float, y: float, z: float) -> Message:
        msg = Message()
        msg.id = 91
        msg.ctrl = 0x03
        msg.params = bytearray(bytes([0x01]))
        msg.params.extend(bytearray(struct.pack("f", x)))
        msg.params.extend(bytearray(struct.pack("f", y)))
        msg.params.extend(bytearray(struct.pack("f", z)))
        msg.params.append(0x00)
        return self._send_command(msg)

    def _set_ptp_joint_params(
        self, v_x: float, v_y: float, v_z: float, v_r: float, a_x: float, a_y: float, a_z: float, a_r: float
    ) -> Message:
        msg = Message()
        msg.id = 80
        msg.ctrl = 0x03
        msg.params = bytearray([])
        msg.params.extend(bytearray(struct.pack("f", v_x)))
        msg.params.extend(bytearray(struct.pack("f", v_y)))
        msg.params.extend(bytearray(struct.pack("f", v_z)))
        msg.params.extend(bytearray(struct.pack("f", v_r)))
        msg.params.extend(bytearray(struct.pack("f", a_x)))
        msg.params.extend(bytearray(struct.pack("f", a_y)))
        msg.params.extend(bytearray(struct.pack("f", a_z)))
        msg.params.extend(bytearray(struct.pack("f", a_r)))
        return self._send_command(msg)

    def _set_ptp_coordinate_params(self, velocity: float, acceleration: float) -> Message:
        msg = Message()
        msg.id = 81
        msg.ctrl = 0x03
        msg.params = bytearray([])
        msg.params.extend(bytearray(struct.pack("f", velocity)))
        msg.params.extend(bytearray(struct.pack("f", velocity)))
        msg.params.extend(bytearray(struct.pack("f", acceleration)))
        msg.params.extend(bytearray(struct.pack("f", acceleration)))
        return self._send_command(msg)

    def _set_ptp_jump_params(self, jump: float, limit: float) -> Message:
        msg = Message()
        msg.id = 82
        msg.ctrl = 0x03
        msg.params = bytearray([])
        msg.params.extend(bytearray(struct.pack("f", jump)))
        msg.params.extend(bytearray(struct.pack("f", limit)))
        return self._send_command(msg)

    def _set_ptp_common_params(self, velocity: float, acceleration: float) -> Message:
        msg = Message()
        msg.id = 83
        msg.ctrl = 0x03
        msg.params = bytearray([])
        msg.params.extend(bytearray(struct.pack("f", velocity)))
        msg.params.extend(bytearray(struct.pack("f", acceleration)))
        return self._send_command(msg)

    def _set_ptp_cmd(self, x: float, y: float, z: float, r: float, mode: MODE_PTP) -> Message:
        msg = Message()
        msg.id = 84
        msg.ctrl = 0x03
        msg.params = bytearray([])
        msg.params.extend(bytearray([mode]))
        msg.params.extend(bytearray(struct.pack("f", x)))
        msg.params.extend(bytearray(struct.pack("f", y)))
        msg.params.extend(bytearray(struct.pack("f", z)))
        msg.params.extend(bytearray(struct.pack("f", r)))
        return self._send_command(msg)

    def _set_end_effector_suction_cup(self, enable: bool = False) -> Message:
        msg = Message()
        msg.id = 62
        msg.ctrl = 0x03
        msg.params = bytearray([])
        msg.params.extend(bytearray([0x01]))
        if enable is True:
            msg.params.extend(bytearray([0x01]))
        else:
            msg.params.extend(bytearray([0x00]))
        return self._send_command(msg)

    def _set_end_effector_gripper(self, enable: bool = False) -> Message:
        msg = Message()
        msg.id = 63
        msg.ctrl = 0x03
        msg.params = bytearray([])
        msg.params.extend(bytearray([0x01]))
        if enable is True:
            msg.params.extend(bytearray([0x01]))
        else:
            msg.params.extend(bytearray([0x00]))
        return self._send_command(msg)

    def _set_end_effector_laser(self, power: int = 255, enable: bool = False) -> Message:
        """Enables the laser.

        Power from 0 to 255.
        """
        msg = Message()
        msg.id = 61
        msg.ctrl = 0x03
        msg.params = bytearray([])
        # msg.params.extend(bytearray([0x01]))
        if enable is True:
            msg.params.extend(bytearray([0x01]))
        else:
            msg.params.extend(bytearray([0x00]))
        # Assuming the last byte is power. Seems to have little effect
        msg.params.extend(bytearray([power]))
        return self._send_command(msg)

    def _set_queued_cmd_start_exec(self) -> Message:
        msg = Message()
        msg.id = 240
        msg.ctrl = 0x01
        return self._send_command(msg)

    def _set_queued_cmd_stop_exec(self) -> Message:
        msg = Message()
        msg.id = 241
        msg.ctrl = 0x01
        return self._send_command(msg)

    def _set_queued_cmd_clear(self) -> Message:
        msg = Message()
        msg.id = 245
        msg.ctrl = 0x01
        return self._send_command(msg)

    def _get_queued_cmd_current_index(self) -> int:
        msg = Message()
        msg.id = 246
        response = self._send_command(msg)
        if response and response.id == 246:
            return self._extract_cmd_index(response)
        else:
            return -1

    @staticmethod
    def _extract_cmd_index(response) -> int:
        return struct.unpack_from("I", response.params, 0)[0]

    def wait_for_cmd(self, cmd_id: int) -> None:
        current_cmd_id = self._get_queued_cmd_current_index()
        while cmd_id > current_cmd_id:
            self.logger.debug("Current-ID", current_cmd_id)
            self.logger.debug("Waiting for", cmd_id)

            current_cmd_id = self._get_queued_cmd_current_index()

    def _set_home_cmd(self) -> Message:
        msg = Message()
        msg.id = 31
        msg.ctrl = 0x03
        msg.params = bytearray([])
        return self._send_command(msg)

    def _set_arc_cmd(
        self, x: float, y: float, z: float, r: float, cir_x: float, cir_y: float, cir_z: float, cir_r: float
    ) -> Message:
        msg = Message()
        msg.id = 101
        msg.ctrl = 0x03
        msg.params = bytearray([])
        msg.params.extend(bytearray(struct.pack("f", cir_x)))
        msg.params.extend(bytearray(struct.pack("f", cir_y)))
        msg.params.extend(bytearray(struct.pack("f", cir_z)))
        msg.params.extend(bytearray(struct.pack("f", cir_r)))
        msg.params.extend(bytearray(struct.pack("f", x)))
        msg.params.extend(bytearray(struct.pack("f", y)))
        msg.params.extend(bytearray(struct.pack("f", z)))
        msg.params.extend(bytearray(struct.pack("f", r)))
        return self._send_command(msg)

    def _set_home_coordinate(self, x: float, y: float, z: float, r: float) -> Message:
        msg = Message()
        msg.id = 30
        msg.ctrl = 0x03
        msg.params = bytearray([])
        msg.params.extend(bytearray(struct.pack("f", x)))
        msg.params.extend(bytearray(struct.pack("f", y)))
        msg.params.extend(bytearray(struct.pack("f", z)))
        msg.params.extend(bytearray(struct.pack("f", r)))
        return self._send_command(msg)

    def _set_jog_coordinate_params(
        self,
        vx: float,
        vy: float,
        vz: float,
        vr: float,
        ax: float = 100,
        ay: float = 100,
        az: float = 100,
        ar: float = 100,
    ) -> Message:
        msg = Message()
        msg.id = 71
        msg.ctrl = 0x03
        msg.params = bytearray([])
        msg.params.extend(bytearray(struct.pack("f", vx)))
        msg.params.extend(bytearray(struct.pack("f", vy)))
        msg.params.extend(bytearray(struct.pack("f", vz)))
        msg.params.extend(bytearray(struct.pack("f", vr)))
        msg.params.extend(bytearray(struct.pack("f", ax)))
        msg.params.extend(bytearray(struct.pack("f", ay)))
        msg.params.extend(bytearray(struct.pack("f", az)))
        msg.params.extend(bytearray(struct.pack("f", ar)))
        return self._send_command(msg)

    def _set_jog_command(self, cmd: int) -> Message:
        msg = Message()
        msg.id = 73
        msg.ctrl = 0x03
        msg.params = bytearray([])
        msg.params.extend(bytearray([0x0]))
        msg.params.extend(bytearray([cmd]))
        return self._send_command(msg)

    def jog_x(self, v: float) -> None:

        self._set_jog_coordinate_params(
            abs(v),
            0,
            0,
            0,
        )
        if v > 0:
            cmd = 1
        elif v < 0:
            cmd = 2
        else:
            cmd = 0

        self.wait_for_cmd(self._extract_cmd_index(self._set_jog_command(cmd)))

    def jog_y(self, v: float) -> None:

        self._set_jog_coordinate_params(0, abs(v), 0, 0)
        if v > 0:
            cmd = 3
        elif v < 0:
            cmd = 4
        else:
            cmd = 0

        self.wait_for_cmd(self._extract_cmd_index(self._set_jog_command(cmd)))

    def jog_z(self, v: float) -> None:

        self._set_jog_coordinate_params(0, 0, abs(v), 0)
        if v > 0:
            cmd = 5
        elif v < 0:
            cmd = 6
        else:
            cmd = 0

        self.wait_for_cmd(self._extract_cmd_index(self._set_jog_command(cmd)))

    def jog_r(self, v: float) -> None:

        self._set_jog_coordinate_params(0, 0, 0, abs(v))
        if v > 0:
            cmd = 7
        elif v < 0:
            cmd = 8
        else:
            cmd = 0

        self.wait_for_cmd(self._extract_cmd_index(self._set_jog_command(cmd)))

    def set_io(self, address: int, state: bool) -> None:

        if not 1 <= address <= 22:
            raise DobotApiException("Invalid address range.")

        msg = Message()
        msg.id = 131
        msg.ctrl = 0x03
        msg.params = bytearray([])
        msg.params.extend(bytearray(struct.pack("B", address)))
        msg.params.extend(bytearray(struct.pack("B", int(state))))

        self.wait_for_cmd(self._extract_cmd_index(self._send_command(msg)))

    def set_hht_trig_output(self, state: bool) -> None:

        msg = Message()
        msg.id = 41
        msg.ctrl = 0x01
        msg.params = bytearray([])
        msg.params.extend(bytearray(struct.pack("B", int(state))))

        self._send_command(msg)

    def get_hht_trig_output(self) -> bool:

        msg = Message()
        msg.id = 41
        msg.ctrl = 0

        response = self._send_command(msg)
        return bool(struct.unpack_from("B", response.params, 0)[0])

    def move_to(self, x: float, y: float, z: float, r: float = 0.0, mode=MODE_PTP.MOVJ_XYZ) -> int:
        return self._extract_cmd_index(self._set_ptp_cmd(x, y, z, r, mode))

    def go_arc(
        self, x: float, y: float, z: float, r: float, cir_x: float, cir_y: float, cir_z: float, cir_r: float
    ) -> int:
        return self._extract_cmd_index(self._set_arc_cmd(x, y, z, r, cir_x, cir_y, cir_z, cir_r))

    def suck(self, enable: bool) -> int:
        return self._extract_cmd_index(self._set_end_effector_suction_cup(enable))

    def set_home(self, x: float, y: float, z: float, r: float = 0.0) -> None:
        self._set_home_coordinate(x, y, z, r)

    def home(self) -> int:
        return self._extract_cmd_index(self._set_home_cmd())

    def grip(self, enable: bool) -> int:
        return self._extract_cmd_index(self._set_end_effector_gripper(enable))

    def laze(self, power: int = 0, enable: bool = False) -> int:
        return self._extract_cmd_index(self._set_end_effector_laser(power, enable))

    def speed(self, velocity: float = 100.0, acceleration: float = 100.0) -> None:
        self.wait_for_cmd(self._extract_cmd_index(self._set_ptp_common_params(velocity, acceleration)))
        self.wait_for_cmd(self._extract_cmd_index(self._set_ptp_coordinate_params(velocity, acceleration)))

    def conveyor_belt(self, speed: float, direction: int = 1, interface: int = 0) -> None:
        if 0.0 <= speed <= 100.0 and (direction == 1 or direction == -1):
            motor_speed = speed * STEP_PER_CIRCLE / MM_PER_CIRCLE * direction
            self._set_stepper_motor(int(motor_speed), interface)
        else:
            raise DobotApiException("Wrong Parameter")

    def _set_stepper_motor(self, speed: int, interface: int = 0, motor_control=True) -> Message:
        msg = Message()
        msg.id = 0x87
        msg.ctrl = 0x03
        msg.params = bytearray([])
        if interface == 1:
            msg.params.extend(bytearray([0x01]))
        else:
            msg.params.extend(bytearray([0x00]))
        if motor_control is True:
            msg.params.extend(bytearray([0x01]))
        else:
            msg.params.extend(bytearray([0x00]))
        msg.params.extend(bytearray(struct.pack("i", speed)))
        return self._send_command(msg)

    def conveyor_belt_distance(self, speed: float, distance: float, direction: int = 1, interface: int = 0) -> None:
        if 0.0 <= speed <= 100.0 and (direction == 1 or direction == -1):
            motor_speed = speed * STEP_PER_CIRCLE / MM_PER_CIRCLE * direction
            steps = (distance * STEP_PER_CIRCLE / MM_PER_CIRCLE) / HUNGARIAN_CONSTANT
            self._set_stepper_motor_distance(int(motor_speed), int(steps), interface)
        else:
            raise DobotApiException("Wrong conveyor belt parameters.")

    def _set_stepper_motor_distance(
        self, speed: int, distance: int, interface: int = 0, motor_control: bool = True
    ) -> Message:
        msg = Message()
        msg.id = 0x88
        msg.ctrl = 0x03
        msg.params = bytearray([])
        if interface == 1:
            msg.params.extend(bytearray([0x01]))
        else:
            msg.params.extend(bytearray([0x00]))
        if motor_control is True:
            msg.params.extend(bytearray([0x01]))
        else:
            msg.params.extend(bytearray([0x00]))
        msg.params.extend(bytearray(struct.pack("i", speed)))
        msg.params.extend(bytearray(struct.pack("I", distance)))
        return self._send_command(msg)

    def _set_cp_params(self, velocity: float, acceleration: float, period: float) -> Message:

        msg = Message()
        msg.id = 90
        msg.ctrl = 0x3
        msg.params = bytearray([])
        msg.params.extend(bytearray(struct.pack("f", acceleration)))
        msg.params.extend(bytearray(struct.pack("f", velocity)))
        msg.params.extend(bytearray(struct.pack("f", period)))
        msg.params.extend(bytearray([0x0]))  # non real-time mode (what does it mean??)
        return self._send_command(msg)

    def _set_cple_cmd(self, x: float, y: float, z: float, power: float, absolute: bool = False) -> Message:

        assert 0 <= power <= 100

        msg = Message()
        msg.id = 92
        msg.ctrl = 0x3
        msg.params = bytearray([int(absolute)])
        msg.params.extend(bytearray(struct.pack("f", x)))
        msg.params.extend(bytearray(struct.pack("f", y)))
        msg.params.extend(bytearray(struct.pack("f", z)))
        msg.params.extend(bytearray(struct.pack("f", power)))
        return self._send_command(msg)

    def engrave(
        self,
        image: np.ndarray,
        pixel_size: float,
        low: float = 0.0,
        high: float = 40.0,
        velocity: float = 5,
        acceleration: float = 5,
        actual_acceleration: float = 5,
    ) -> None:
        """Shade engrave the given image.

        :param image: NumPy array representing the image. Should be 8 bit grayscale image.
        :param pixel_size: Pixel size in mm.
        :param low: Image values will be scaled to range of <low, high>.
        :param high: dtto
        :param velocity: Maximum junction velocity (CPParams).
        :param acceleration: Maximum planned accelerations (CPParams).
        :param actual_acceleration: Maximum actual acceleration, used in non-real-time mode.
        :return:

        Example usage:

        >>> from PIL import Image
        >>> import numpy as np
        >>> d = DobotApi()
        >>> im = Image.open("image.jpg")
        >>> im = im.convert("L")
        >>> im = np.array(im)

        >>> x, y = d.get_pose().position[0:2]
        >>> d.wait_for_cmd(d.move_to(x, y, -74.0))

        >>> d.engrave(im, 0.1)
        """

        image = image.astype("float64")
        image = 255.0 - image
        image = (image - image.min()) / (image.max() - image.min()) * (high - low) + low

        x, y, z = self.get_pose().position[0:3]  # get current/starting position

        self.wait_for_cmd(self.laze(0, False))
        self._set_queued_cmd_clear()
        self.wait_for_cmd(self._extract_cmd_index(self._set_cp_params(velocity, acceleration, actual_acceleration)))

        self._set_queued_cmd_stop_exec()
        stopped = True

        indexes: deque[int] = deque()

        for row_idx, row in enumerate(image):

            # first feed the queue to be almost full
            if stopped and len(indexes) > MAX_QUEUE_LEN - 2:
                self._set_queued_cmd_start_exec()
                stopped = False

            if row_idx % 2 == 1:
                data = reversed(row)
                rev = True
            else:
                data = row
                rev = False

            for col_idx, ld in enumerate(data):

                if not rev:
                    y_ofs = col_idx * pixel_size
                else:
                    y_ofs = (len(row) - 1 - col_idx) * pixel_size

                indexes.append(
                    self._extract_cmd_index(self._set_cple_cmd(x + row_idx * pixel_size, y + y_ofs, z, ld, True))
                )

                # then feed it as necessary to keep it almost full
                while not stopped and len(indexes) > MAX_QUEUE_LEN - 12:
                    self.wait_for_cmd(indexes.popleft())

        self.wait_for_cmd(self.laze(0, False))
